/* 
 * Program:             $RCSfile: PIDB.h,v $  $Revision: 31.24 $
 * 
 * Purpose:
 * 
 * Author:              T.Mimura  1993/12/01
 *                      S.Yamamoto  1994/09/01
 * 
 * (C) Copyright:       T.Mimura and S.Yamamoto  1993 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef _SPD_PIDB_H
#define _SPD_PIDB_H

#define PIDB_TUPLE_SEPARATOR            '\2'
#define IS_PIDB_TUPLE_SEPARATOR(x)      ((x) == PIDB_TUPLE_SEPARATOR)

#define PIDB_LINE_SEPARATOR             '\n'
#define IS_PIDB_LINE_SEPARATOR(x)       ((x) == PIDB_LINE_SEPARATOR)

#define CPPinfoN    CPPinfoMAX      /* Number of files for CPPinfo. */

enum CPPinfo {
    CPPinfoDEFINE = 0,
    CPPinfoEXPAND,
    CPPinfoIFEXPR,
    CPPinfoMAX                      /* Max value of enumeration. */
};

#define PIDB_USE_FILENAME_INDEX
/* 
 *  PIDB_USE_FILENAME_INDEX is to improve performance of libPIDB.
 *
 *  PIDB_USE_FILENAME_INDEX changes the comparison method of file names
 *  in 'equalOutputText()' and 'equalInputText()' from strcmp() to 
 *  equality of file name indexes. It adds a member 'fname_idx' in structures 
 *  MacroExpandRegion and MacroCallRegion. Because the member 'fname_idx' 
 *  can not be set automatically, clients of libPIDB must set 'fname_idx' of 
 *  each instance of structures by calling the function 'pidbConvFileId()'.
 *  'pidbConvFileId()' returns the object id of the file specified in its
 *  argument.
 *
 *  Added by A.Yoshida, 2002/07/22.
*/

/******************************************************************************/

/*
 * Macro definition information.
 */

/*
 * ޥˡ
 */

typedef enum    {
    DEFINE_INITIALIZATION = 0,      /* Predifined. */
    DEFINE_D_OPTION,                /* "-D" option. */
    DEFINE_FILE                     /* Defined with #define directive. */
} DefineType;

/*
 * ޥΰ֡
 */

typedef struct  {
    DefineType  type;
    /* ʲΥФ(type == DEFINE_FILE)ΤȤˤΤ̣߰ġ */
    char    *fname;                 /* ե̾ */
    int     start;                  /* γϰ֡ */
    int     end;                    /* νλ֡ */
} DefineRegion;
    
/*
 * ޥŸˤλȡ
 */

typedef struct  {
    int         nchar;          /* ncharʸФƤargn */
    int         argn;
    SpdBoolean  raw_before;     /* "##"뤫 */
    SpdBoolean  raw_after;      /* "##"뤫 */
    SpdBoolean  stringify;      /* "#"뤫 */
} ArgReference;

/*
 * ޥ˴ؤ
 */

typedef struct {
    int     ID;         /* ID */
    DefineRegion pos;   /* 줿֡ */
    char    *name;      /* ޥ̾ */
    int     narg;       /* θĿ(Ȥʤ-1,0ξ⤢) */
    char    *expansion; /* ִʸ(ʬ) */
    int     nref;       /* Ÿκݤ˲뤫 */
    ArgReference *ref;  /* (refĤǤ) */
    char    **arg;      /* ʬΰ̾(nargĤǤ) */
    int     output;     /* ϥեб֡ */
} DEFINEinfo;

/******************************************************************************/

/*
 * Macro expansion information.
 */

/*
 * ޥθƤӽФΰ֤μࡥ
 */

typedef enum    {
    CALL_FILE = 0,          /* եǡ */
    CALL_MACROEXPANSION,    /* ޥŸǡ */
    CALL_MACROARG           /* ޥƤӽФΰǡ */
} MacroCallType;

/*
 * ޥθƤӽФΰ(ϰ)
 */

typedef struct  {
    MacroCallType   type;   /* ƤӽФ֤μࡥ */
    int         start;      /* Ƭ֡ */
    int         end;        /* λ֡ */
    char        *fname;     /* ե̾(type == CALL_FILEξ) */

#ifdef PIDB_USE_FILENAME_INDEX
    SpdObjId    fname_idx;  /* Index of file name. */
#endif

    int             macexpID;

    /*
     * ƤӽФΥޥŸȤμֹ.
     * macexp == expandInfo[macexpID]Ȥطˤʤ
     * (type == CALL_MACROEXPANSION || type == CALL_MACROARGξΤ)
     */
    struct expand_info  *macexp;

    int         argn;       /* ֹ(type == CALL_MACROARGξ) */
} MacroCallRegion;

/*
 * ޥθƤӽФΰ(ϰ)륭塼
 */

typedef struct  {
    MacroCallRegion *buf;
    int         extent;
    int         top, tail;
    int         used;
} MacroCallRegionQueue;

/*
 * ޥŸ̤νμࡥ
 */

typedef enum    {
    EXPAND_FILE = 0,        /* ϥե롥 */
    EXPAND_MACROARG         /* ޥΰ */
} MacroExpandType;

/*
 * ޥŸ̤ΰ֡
 */

typedef struct {
    MacroExpandType type;
    int         start;      /* Ƭ֡ */
    int         end;        /* λ֡ */
    SpdConstString fname;     /* ե̾(type == EXPAND_FILEξ) */

#ifdef PIDB_USE_FILENAME_INDEX
    SpdObjId    fname_idx;  /* Index of file name. */
#endif

    int         macexpID;

    /*
     * ˤʤĥޥŸȤμֹ桥
     * macexp == expandInfo[macexpID]Ȥطˤʤ(type == MACRO_ARGξ)
     */
    struct expand_info  *macexp;

    int         argn;       /* ֹ(type == EXPAND_MACROARGξ) */
} MacroExpandRegion;

/*
 * ޥŸ̤ΰ(ϰ)륭塼
 */

typedef struct  {
    MacroExpandRegion   *buf;
    int         extent;
    int         top, tail;
    int         used;
} MacroExpandRegionQueue;

/*
 * ޥŸθġΰ˴ؤ
 */

typedef struct  {
    int     start;                  /* ƽФǤϰϡ */
    int     end;
    int     raw_length;             /* Ĺ */
    int     expanded_length;        /* ŸĹ */
    int     stringified_length;     /* ʸ󲽸Ĺ */
    int     trailing_space;         /* ƬϢ³ʸĹ */
    int     following_space;        /* Ϣ³ʸĹ */
} MacroArgInfo;

/*
 * ޥŸ˴ؤ
 */

typedef struct  expand_info {
    int         ID;
    int         macID;          /* ƤӽФ줿ޥID. */
    DEFINEinfo      *mac;       /* ƤӽФ줿ޥ˴ؤ */
    MacroCallRegion input;      /* ޥƤӽФ֡ */
    MacroArgInfo    *arg;       /* ưξ(mac->nargĤǤ) */
    MacroExpandRegion   output; /* Ÿ̤νϰ֡ */
} EXPANDinfo;

typedef struct ifexpr_info {
    char        *cfilename;
    int         clineno;
    SpdBoolean  ifexpr;
    char        *ifilename;
    int         ilineno;
} IFEXPRinfo;

/******************************************************************************/

extern DEFINEinfo   *defineInfo;
extern int          defineInfoN;
extern EXPANDinfo   *expandInfo;
extern int          expandInfoN;
extern IFEXPRinfo   *ifexprInfo;
extern int          ifexprInfoN;

extern  SpdBoolean  CPPinfoVerbose;
extern  int         CPPinfoQueueExtent;

/******************************************************************************/

/* "streamCPPinfo.c" */
void    setCPPinfoPath(char *path);
int     spdInitializePidbCache(SpdObjId ifileId);
void    spdReleasePidbCache(void);

/* printCPPinfo.c */
void CPPinfoMessage(const char *format, ...);
void printDEFINEinfo(FILE *fp, int n);
void printEXPANDinfo(FILE *fp, int n);

/* mqueueCPPinfo.c */
void createMacroCallRegionQueue(MacroCallRegionQueue *queue, int extent);
void freeMacroCallRegionQueue(MacroCallRegionQueue *queue);
void pushMacroCallRegionQueue(MacroCallRegionQueue *queue, MacroCallRegion reg);
MacroCallRegion popMacroCallRegionQueue(MacroCallRegionQueue *queue);
int usedMacroCallRegionQueue(MacroCallRegionQueue *queue);

void createMacroExpandRegionQueue(MacroExpandRegionQueue *queue, int extent);
void freeMacroExpandRegionQueue(MacroExpandRegionQueue *queue);
void pushMacroExpandRegionQueue(MacroExpandRegionQueue *queue,
                                MacroExpandRegion reg);
MacroExpandRegion popMacroExpandRegionQueue(MacroExpandRegionQueue *queue);
int usedMacroExpandRegionQueue(MacroExpandRegionQueue *queue);

/* macroCPPinfo.h */
SpdBoolean equalInputText(MacroCallRegion *a, MacroCallRegion *b);
int isInMacroCall(MacroCallRegion *t);
/* int isInMacroArg(int id, MacroCallRegion *t); */
SpdBoolean isMacroCallRedionIncluded(MacroCallRegion *a, MacroCallRegion *b);
/* 
 * int macroCallToExpand(int id, MacroCallRegion *target,
 *                       MacroExpandRegionQueue *eq);
 */
int completeMacroExpand(MacroCallRegion *target,
                        MacroExpandRegionQueue *eq);

/* revmacroCPPinfo.c */
SpdBoolean equalOutputText(MacroExpandRegion *a, MacroExpandRegion *b);
int isInMacroExpand(MacroExpandRegion *exRegion, int start, int end);
int isInMacroExpandArg(int id, MacroExpandRegion *exRegion, int *n);
int reverseParseAnOutputText(MacroExpandRegion *target,
                             MacroCallRegion *result);

/* cpplineconv.c */
int lineConvertCtoI(char *name, int cline, char *ibuf, int ilen);
int lineConvertItoC(char *name, int iline, char *ibuf, int ilen);

/* cppcharconv.c */
int cppCharNotInMacroItoC(SpdConstString iname, int ichar, char *ibuf,
                          int ilen, char *cname);
int cppCharConvCtoI(char *cname, int offset, MacroExpandRegionQueue *queue,
                    char *cbuf, int clen, char *iname, char *ibuf, int ilen);
int cppCharConvItoC(SpdConstString iname, int offset, char *ibuf, int ilen,
                    char *cname);
    
/* PIDB_stream.c */
void pidbWriteLineSeparator(FILE *fp);
void pidbWriteTupleSeparator(FILE *fp);

/* PIDB.c */
#ifdef PIDB_USE_FILENAME_INDEX
SpdObjId pidbConvFileId(SpdConstString fname);
void pidbConvFileId_free(void);
#endif

/******************************************************************************/

/*
 *                      Line table.
 */

#if defined(SAPID_AR3_KERNEL)
#include <Sapid/lineTbl.h> 
#endif

#endif /* _SPD_PIDB_H */
