/* 
 * Program:             $RCSfile: SMU.h,v $  $Revision: 30.32 $
 * 
 * Purpose:
 * 
 * Author:              A.Watanabe  1996/05/15
 *                      S.Yamamoto  1996/06/05
 * 
 * (C) Copyright:       A.Watanabe and S.Yamamoto  1996 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef _SMU_H
#define _SMU_H

/* Data type. */

typedef struct _SpdObjIdArray {
    size_t      size;
    SpdObjId    *id;
} SpdObjIdArray;

typedef struct _SpdRelIdArray {
    size_t      size;
    SpdRelId    *id;
} SpdRelIdArray;

typedef struct _SpdRelArray {
    size_t      size;
    SpdRel      *rel;
} SpdRelArray;

typedef struct _SpdOccArray {
    size_t      size;
    SpdOcc      *occ;
} SpdOccArray;

typedef SpdBoolean (*SpdCondObjIdFunc)(SpdObjId);
typedef SpdBoolean (*SpdCondRelFunc)(SpdRel);
typedef SpdBoolean (*SpdCondRelIdFunc)(SpdRelId);
typedef SpdBoolean (*SpdCondOccFunc)(SpdOcc);

typedef int (*SpdCompareObjIdFunc)(const SpdObjId *, const SpdObjId *);
typedef int (*SpdCompareRelFunc)(const SpdRel *, const SpdRel *);
typedef int (*SpdCompareRelIdFunc)(const SpdRelId *, const SpdRelId *);
typedef int (*SpdCompareOccFunc)(const SpdOcc *, const SpdOcc *);

/* Prototypes. */

void spdFreeObjIdArray(SpdObjIdArray array);
void spdFreeRelIdArray(SpdRelIdArray array);
void spdFreeRelArray(SpdRelArray array);
void spdFreeOccArray(SpdOccArray array);

SpdObjIdArray spdGetObjIdArray(SpdConstString className,
                               SpdConstString name,
                               SpdCondObjIdFunc cond);
SpdObjIdArray spdGetObjIdSortedArray(SpdConstString className,
                                     SpdConstString name,
                                     SpdCondObjIdFunc cond,
                                     SpdCompareObjIdFunc compar);

SpdRelArray spdGetRelArray(SpdObjId objectId,
                           SpdConstString relationName,
                           SpdConstString attrName,
                           SpdCondRelFunc cond);
SpdRelArray spdGetRelSortedArray(SpdObjId objectId,
                                 SpdConstString relationName,
                                 SpdConstString attrName,
                                 SpdCondRelFunc cond,
                                 SpdCompareRelFunc compar);
SpdObjIdArray spdGetRelObjArray(SpdObjId objectId,
                                SpdConstString relationName,
                                SpdConstString attrName,
                                SpdCondObjIdFunc cond);
SpdObjIdArray spdGetRelObjSortedArray(SpdObjId objectId,
                                      SpdConstString relationName,
                                      SpdConstString attrName,
                                      SpdCondObjIdFunc cond,
                                      SpdCompareObjIdFunc compar);
SpdRelIdArray spdGetRelIdArray(SpdObjId objectId1,
                               SpdObjId objectId2,
                               SpdConstString relationName,
                               SpdCondRelIdFunc cond);
SpdRelIdArray spdGetRelIdSortedArray(SpdObjId objectId1,
                                     SpdObjId objectId2,
                                     SpdConstString relationName,
                                     SpdCondRelIdFunc cond,
                                     SpdCompareRelIdFunc compar);
SpdOccArray spdGetOffsetOfObjArray(SpdObjId objectId,
                                   SpdCondOccFunc cond);
SpdOccArray spdGetOffsetOfObjSortedArray(SpdObjId objectId,
                                         SpdCondOccFunc cond,
                                         SpdCompareOccFunc compar);
SpdOccArray spdGetIncludedOccArray(SpdObjId objectId,
                                   SpdConstString objectType,
                                   SpdObjId fileId,
                                   SpdCondOccFunc cond);
SpdOccArray spdGetIncludedOccSortedArray(SpdObjId objectId,
                                         SpdConstString objectType,
                                         SpdObjId fileId,
                                         SpdCondOccFunc cond,
                                         SpdCompareOccFunc compar);

/* Sample compare() functions. */

int spdCompareObjIdName(const SpdObjId *a, const SpdObjId *b);
int spdCompareRelOffset(const SpdRel *a, const SpdRel *b);
int spdCompareRelIdOffset(const SpdRelId *a, const SpdRelId *b);
int spdCompareOccOffset(const SpdOcc *a, const SpdOcc *b);
int spdCompareOccName(const SpdOcc *a, const SpdOcc *b);

/* Sample cond() functions. */

SpdBoolean spdObjIdIsDeclNormal(SpdObjId id);
SpdBoolean spdObjIdIsDeclFuncdef(SpdObjId id);
SpdBoolean spdObjIdIsDeclEnumerator(SpdObjId id);
SpdBoolean spdObjIdIsDeclParameter(SpdObjId id);
SpdBoolean spdObjIdIsDeclMember(SpdObjId id);
SpdBoolean spdObjIdIsDeclType(SpdObjId id);

SpdBoolean spdObjIdIsIdVariable(SpdObjId id);
SpdBoolean spdObjIdIsIdFunction(SpdObjId id);
SpdBoolean spdObjIdIsIdMember(SpdObjId id);
SpdBoolean spdObjIdIsIdEnumerator(SpdObjId id);
SpdBoolean spdObjIdIsIdTypename(SpdObjId id);
SpdBoolean spdObjIdIsIdTag(SpdObjId id);

/******************************************************************************/

/* Prototypes in "spdGetProgramName.c". */

SpdString   smuGetAProgramName(void);
SpdBoolean  smuIsValidSDB(void);

/* Prototypes in "predVariable.c". */

SpdBoolean  smuIsArgument(SpdObjId ident_id);
SpdBoolean  smuIsGlobalVariable(SpdObjId ident_id);
SpdBoolean  smuIsLocalVariable(SpdObjId ident_id);
SpdBoolean  smuIsStaticId(SpdObjId ident_id);

/* In "xrefTable.c". */

typedef struct _SmuXrefTableElem {
    SpdObjId        callerId;
    SpdObjId        calleeId;
    SpdRelId        relationId;
} SmuXrefTableElem;

typedef struct _SmuXrefTable {
    size_t              size;
    SmuXrefTableElem    *elem;
} SmuXrefTable;

typedef struct _SmuXrefTables {
    SmuXrefTable    funcCallTable;
    SmuXrefTable    variableTable;
    SmuXrefTable    globalVariableTable;
    SmuXrefTable    localVariableTable;
    SmuXrefTable    macroCallTable;
} SmuXrefTables;

typedef enum {smuXrefEachOcc = 0,  smuXrefEachObj} SmuXrefFlag;

SmuXrefTables   smuMakeXrefTables(SpdObjId prog_id, SmuXrefFlag flag);
void            smuFreeXrefTables(SmuXrefTables tables);

/******************************************************************************/

/* Prototypes in "smuGetAttr.c". */

SpdBoolean  smuGetObjAttr(SpdObjId objId, SpdString attrName,
                          SpdString *retAttr, SpdString *retReadable);
SpdBoolean smuGetClassAttr(int classId, SpdString attrName, int attrVal,
                           SpdString *retAttr, SpdString *retReadable);

SpdString smuGetRelParentAttrName(SpdRelId relId);
SpdString smuGetRelChildAttrName(SpdRelId relId);

/******************************************************************************/

/* Prototypes in "exprUtil.c", "exprType.c" and "typeUtil.c". */

#if defined(SAPID_AR4_KERNEL)
SpdObjId smuGetExprType(SpdObjId exprid);
#endif

int smuGetExprSort(SpdObjId exprId);

SpdBoolean smuObjIdIsTypePointer(SpdObjId id);
SpdBoolean smuObjIdIsTypeIntegral(SpdObjId id);
SpdBoolean smuObjIdIsTypeDouble(SpdObjId id);
SpdBoolean smuObjIdIsTypeLong(SpdObjId id);
SpdBoolean smuObjIdIsTypeUnsigned(SpdObjId id);

SpdObjId smuFindType(int qualifier, int sign, int size, int type);
SpdObjId smuFindTypedef(SpdString name);
SpdObjId smuFindPointerTo(SpdObjId type);
SpdObjId smuFindArrayOf(SpdObjId type);

/******************************************************************************/

/* Constants and prototypes in "operatorUtil.c". */

/* ѤĤ: ޤ黻 */
#define OPCATEGORY_ASSIGN       (1 << 0) 

/* ѷƤŬѤǤ뤫
   OPCATEGORY_INT  OPCATEGORY_ARITH */
#define OPCATEGORY_ARITH        (1 << 1)

/* intΤߤŬѤǤ뤫 */
#define OPCATEGORY_INT          (1 << 2)

/* ݥ󥿷ŬѤǤ뤫 */
#define OPCATEGORY_POINTER      (1 << 3)

/* ͤΤߤŬѤǤ뤫: ʳ */
#define OPCATEGORY_LVALUE       (1 << 4)

/* 򴹲ǽ */
#define OPCATEGORY_COMMUTATIVE  (1 << 5)

/* 󥹥ݥȤĤ */
#define OPCATEGORY_SEQUENCEP    (1 << 6)

/* ¦μΤƤ뤫:
   OP_ASSIGN (=), OP_COMMA (,)ʤɤκ¦μϻȤʤ */
#define OPCATEGORY_TRASHLVALUE  (1 << 7)

/* ¦ηΤƤ뤫:
   OP_ASSIGN (=), OP_SHL (<<)ʤɤα¦μηϻȤʤ */
#define OPCATEGORY_TRASHRTYPE   (1 << 8)

/* 黻Ҥ: ӤANDʤɡͤintˤʤ */
#define OPCATEGORY_LOGICAL      (1 << 9)

int smuGetOperatorTerms(SpdObjId op);
int smuGetOperatorCategory(SpdObjId op);
int smuInitOperatorCategory(void);

/******************************************************************************/

/* Prototypes in "smuFileUtil.c". */

FILE *smuOpenSrcFile(SpdObjId fileId);

SpdBoolean  smuIsSystemFileName(SpdString file);
SpdBoolean  smuIsSystemFile(SpdObjId pfileId);
SpdBoolean  smuIsSystemObject(SpdObjId objectId);

int spdGetFileLength(SpdObjId fileId);
int spdGetFileSize(SpdObjId fileId);
int spdGetFileLengthByName(SpdConstString fileName);

int spdReadFileInBuf(SpdObjId fileId, char **buf);
int spdReadFileInBufByName(SpdConstString fname, char **buf);
int spdReadJavaInBufByName(SpdConstString fileName, char **buf);

void smuSetSystemPath(char *path_list);

/******************************************************************************/

/* 
 * Moved from SpdUtil by H.Ohhashi 1999/11/25.
 * Prototypes in "spdConvFileId.c".
 */

SpdObjId spdConvFileIdItoC(SpdObjId ifileId);
SpdObjId spdConvFileIdCtoI(SpdObjId cfileId);
SpdObjId spdConvIdPfileToFile(SpdObjId pfileId);
SpdObjId smuGetFileIdByName(SpdConstString fileName);

/******************************************************************************/

/* Prototypes in "smuSetSDBinfo.c". */

void smuSetSDBinfo(SpdInfo *info);

/******************************************************************************/

/* Traverse funcions. */
#include <Sapid/spdTraverse.h>

/* Declarative APIs. */
#include <Sapid/spdDeclarativeAPI.h>

#endif /* _SMU_H */
