/* 
 * Program:             $RCSfile: Sapid.h,v $  $Revision: 30.164 $
 * 
 * Purpose:             Common header of the project Sapid.
 * 
 * Author:              K.Mabuchi  1991/09/30
 *                      S.Yamamoto  1994/04/01
 * Modified             H.Ariga  1993/07/01
 *                      T.Mimura  1994/10/05
 * 
 * (C) Copyright:       K.Mabuchi and S.Yamamoto  1991 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Sapid.h,v 30.164 2003/08/16 12:19:21 yamamoto Exp yamamoto $
 */ 

/* 
 * Include "stdio.h" before this file "Sapid.h" is included.
 */

/* 
 * ToDo:
 * 
 * Should change macro prefix "SAPID_" into "SPD_" (S.Yamamoto 1999/04/17).
 */

#ifndef _SAPID_H
#define _SAPID_H

/* #define NDEBUG */
#include <assert.h>

#include <Sapid/version.h>

/******************************************************************************/

/*
 *                      Configuration.
 */

/* 
 * Dependency diagram of AR[34]:
 * 
 * SAPID_AR4 <--- SAPID_AR4_KERNEL
 *     |                 |
 *     v                 v
 * SAPID_AR3 <--- SAPID_AR3_KERNEL
 */

/* SAPID_AR4_KERNEL requires SAPID_AR3_KERNEL. */

#if defined(SAPID_AR4_KERNEL) && !defined(SAPID_AR3_KERNEL)
#define SAPID_AR3_KERNEL
#endif

/* SAPID_AR[34]_KERNEL requires SAPID_AR[34]. */

#if defined(SAPID_AR3_KERNEL) && !defined(SAPID_AR3)
#define SAPID_AR3
#endif

#if defined(SAPID_AR4_KERNEL) && !defined(SAPID_AR4)
#define SAPID_AR4
#endif

/* SAPID_AR4 requires SAPID_AR3. */

#if defined(SAPID_AR4) && !defined(SAPID_AR3)
#define SAPID_AR3
#endif

/******************************************************************************/

/* 
 * Dependency diagram of AR[35]:
 * 
 * SAPID_AR3 <--- SAPID_AR3_KERNEL
 *     ^                 ^
 *     |                 |
 * SAPID_AR5 <--- SAPID_AR5_KERNEL
 */

/* SAPID_AR5_KERNEL requires SAPID_AR3_KERNEL. */

#if defined(SAPID_AR5_KERNEL) && !defined(SAPID_AR3_KERNEL)
#define SAPID_AR3_KERNEL
#endif

/* SAPID_AR5_KERNEL requires SAPID_AR5. */

#if defined(SAPID_AR5_KERNEL) && !defined(SAPID_AR5)
#define SAPID_AR5
#endif

/* SAPID_AR5 requires SAPID_AR3. */

#if defined(SAPID_AR5) && !defined(SAPID_AR3)
#define SAPID_AR3
#endif

/******************************************************************************/

/* Configuration check. */

#if defined(SAPID_AR1) || defined(SAPID_AR2)
    "Obsolete configuration."
#endif

#if defined(SAPID_AR4) && defined(SAPID_AR5)
    "Invalid configuration."
#endif

/* 
 * #if !defined(SAPID_AR4) && !defined(SAPID_AR5)
 *     "Invalid configuration."
 * #endif
 */

#if defined(USE_SPDSET1) && defined(USE_SPDSET3)
    "Invalid configuration."
#endif

/******************************************************************************/

/*
 * SAPID_MEMORY_PROFILE is used in:
 * "<SAPID_TOP>/analyzer/sdbin4/sdbin4.c".
 * "<SAPID_TOP>/tool/SPIE/spieSub.c".
 */
/* #define SAPID_MEMORY_PROFILE */

/*
 * SAPID_READ_ROUTINE_MEMORY_PROFILE is used in:
 * "<SAPID_TOP>/AR4/AR4/read_sdb.c".
 */
/* #define SAPID_READ_ROUTINE_MEMORY_PROFILE */

/*
 * SAPID_SAFETY_MODE_0:
 * Checking validity of attribute names in functions
 * spdGetAttrVal{Int,String}() and spdGetAttrValIntOf{Obj,Rel}().
 */
/* #define SAPID_SAFETY_MODE_0 */

/*
 * SAPID_SAFETY_MODE_1:
 * Checking validity of object and relation id in functions
 * spdGetAttrVal{Int,String}() and spdGetAttrValIntOf{Obj,Rel}().
 */
/* #define SAPID_SAFETY_MODE_1 */

/******************************************************************************/

/*
 *                      For fixing known bugs.
 */

#include <Sapid/SapidBug.h>

/******************************************************************************/

/*
 *                      Portability.
 */

/*
 * Headers.
 */

/* "pwd.h". */

#if !defined(HAVE_PWD_H) && !defined(WIN32)
#define HAVE_PWD_H
#endif

/* "stddef.h". Just for a macro offsetof(). */

#if !defined(HAVE_STDDEF_H) && (defined(WIN32) || defined(__CYGWIN__))
#define HAVE_STDDEF_H
#endif

/* "sys/time.h" for gettimeofday(). */

#if !defined(HAVE_SYS_TIME_H) && !defined(WIN32)
#define HAVE_SYS_TIME_H
#endif

/* "sys/times.h". */

#if !defined(HAVE_SYS_TIMES_H) && !defined(WIN32)
#define HAVE_SYS_TIMES_H
#endif

/* "unistd.h". */

#if !defined(HAVE_UNISTD_H) && \
    (defined(__hpux) || defined(__sgi) || \
     defined(__FreeBSD__) || defined(__linux__))
#define HAVE_UNISTD_H
#endif

/*
 * Library functions.
 */

/* Header of alloca(). */

#if !defined(HAVE_ALLOCA_H) && \
    !defined(__GNUC__) && (defined(__sgi) || defined(__sun))
#define HAVE_ALLOCA_H
#endif

/* htonl() and ntohl(). */

#if defined(__linux__)
#include <netinet/in.h>
#endif

/* isatty(). */

#if !defined(HAVE_ISATTY) && \
    (defined(__hpux) || defined(__sgi) || defined(__sun) || \
     defined(__FreeBSD__))
#define HAVE_ISATTY
#endif

/* sbrk(). */

#if !defined(HAVE_SBRK) && \
    (defined(__hpux) || defined(__sgi) || defined(__sun) || \
     defined(__FreeBSD__) || defined(__linux__) || defined(__CYGWIN__))
#define HAVE_SBRK
#endif

/* sysconf(). */

#if !defined(HAVE_SYSCONF) && \
    defined(HAVE_UNISTD_H) && (defined(__FreeBSD__) || defined(__linux__)) 
#define HAVE_SYSCONF
#endif

/* times(). */

#if defined(HAVE_SYS_TIMES_H)
#define HAVE_TIMES
#endif

/*
 * Path consists of [<driveName>:]<path>.
 * Use CR + NL as line separator.
 */

#if defined(__CYGWIN__) || defined(WIN32)
#define SPD_ON_WINDOWS_FS
#endif

/******************************************************************************/

/*
 *                      To make Sapid portable.
 */

#if !defined(SAPID_SML)

/*
 * Don't include "X11/{Xos.h,Xfuncs.h}" if you use VC++ and nmake.
 */
/* 
 * #if !defined(WIN32)
 * #define SPD_USE_X_HEADERS
 * #endif
 */
#define SPD_USE_X_HEADERS

/*
 * For strchr(), strrchr() and strerror(). For struct timeval.
 */
#if defined(SPD_USE_X_HEADERS)
#include <X11/Xos.h>
#else
#include <string.h>
#endif

/*
 * For atexit().
 * Suggested by M.Araki 1998/04/25.
 */
#ifdef SUNOS_4
#define atexit(f)       on_exit((f), 0)
#endif

/*
 * For bcopy(), bzero() and bcmp().
 * Don't use these functions, because they are not in ANSI C library.
 */
#if defined(SPD_USE_X_HEADERS) && !defined(__sun)
#include <X11/Xfuncs.h>
#endif

#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#endif

#if defined(X_NOT_STDC_ENV) || (defined(__mips) && !defined(__sgi))
char    *getenv();
char    *malloc(), *calloc(), *realloc();
#endif

#endif /* !defined(SAPID_SML) */

/*
 * For getopt().
 * Library getopt() is declared in various header files.
 * <unistd.h>: HP-UX.
 * <stdlib.h>: IRIX and SunOS.
 * <unistd.h>: Linux with __USE_POSIX2.
 * <getopt.h>: Linux without __USE_POSIX2 and Cygwin B19.
 */

#if defined(__CYGWIN__) || (defined(__linux__) && !defined(__USE_POSIX2))

#include <getopt.h>

#elif defined(SUNOS_4) || defined(WIN32)

extern int      getopt(int, char * const *, const char *);

extern char     *optarg;
extern int      optind, opterr;

#endif /* defined(SUNOS_4) || defined(WIN32) */

#if !defined(WIN32)
#define SPD_PATH_SEP        '/'
#define SPD_PATH_SEP_STR    "/"
#else
#define SPD_PATH_SEP        '\\'
#define SPD_PATH_SEP_STR    "\\"
#endif

/*
 * The maximum number of numerals that result when a 64-bit
 * integer is converted to a string, plus one for a trailing null byte,
 * plus one for a sign. 
 */

#define SPD_MAX_INT_LEN         21

#define SPD_XML_ATTR_LEN        1024

/******************************************************************************/

/*
 *                      Basic type definitions.
 */

/*
 * Note:
 * 
 * Macro SPD_NON_ID is used as special return value of various functions.
 * For example:
 *
 * while ((objId = spdGetRelObj()) != SPD_NON_ID) {
 *     ...
 * }
 *
 * SPD_NON_ID never conflict with any valid Sapid object id.
 * Don't assume SPD_NON_ID is always defined as -1 and is negative integer.
 * It would be defined as 0 in the future.
 */

#if defined(USE_UNSIGNED_ID)

#define SPD_NON_DB_ID   INT_MAX
#define SPD_NON_MID     INT_MAX
#define SPD_NON_CID     INT_MAX
#define SPD_NON_ID      INT_MAX

typedef unsigned int    SpdDBId;
typedef unsigned int    SpdModelId;
typedef unsigned int    SpdClassId;
typedef unsigned int    SpdObjId;
typedef unsigned int    SpdRelId;

#else /* !defined(USE_UNSIGNED_ID) */

#define SPD_NON_DB_ID   (-1)
#define SPD_NON_MID     (-1)
#define SPD_NON_CID     (-1)
#define SPD_NON_ID      (-1)

typedef int             SpdDBId;
typedef int             SpdModelId;
typedef int             SpdClassId;
typedef int             SpdObjId;
typedef int             SpdRelId;

#endif /* !defined(USE_UNSIGNED_ID) */

/******************************************************************************/

/*
 *                      Basic type definitions.
 */

typedef int         SpdBoolean;

/* Should be "unsigned char"? */
typedef char        *SpdString;

typedef const char  *SpdConstString;

/* Super class of SpdObjId and SpdRelId. */
typedef SpdObjId    SpdId;

typedef int         SpdAttrId;

/******************************************************************************/

/*
 *                      Private type definitions.
 */

#ifdef SAPID_AR4_KERNEL

/* For only spdGetAttrIdByName3(). */
typedef int         SpdAttrNameId;

#endif /* SAPID_AR4_KERNEL */

/******************************************************************************/

/*
 *                      Basic constants.
 */

#ifndef TRUE
#define TRUE                    1
#endif

#ifndef FALSE
#define FALSE                   0
#endif

#ifndef ON
#define ON                      1
#endif

#ifndef OFF
#define OFF                     0
#endif

#ifndef EXIT_SUCCESS    
#define EXIT_SUCCESS            0
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE            1
#endif

/*
 * The constant FILENAME_MAX is usually defined in "stdio.h".
 * Include "stdio.h" before "Sapid.h" is included.
 */

#ifndef FILENAME_MAX
#define FILENAME_MAX            1024
#endif

/*
 * Use MAXPATHLEN instead of FILENAME_MAX, if value of macro
 * FILENAME_MAX is too short.
 * FILENAME_MAX is defined as 14 on HP-UX 10.01.
 */

#if defined(__hpux) && (FILENAME_MAX < 16)

#include <sys/param.h>          /* For MAXPATHLEN. */

#if (FILENAME_MAX < MAXPATHLEN)

#undef FILENAME_MAX
#define FILENAME_MAX            MAXPATHLEN

#endif /* (FILENAME_MAX < MAXPATHLEN) */

#endif /* defined(__hpux) && (FILENAME_MAX < 16) */

/*
 * These constants are usually defined in "unistd.h".
 */

#if defined(__mips) && !defined(SEEK_SET)
#define SEEK_SET    0   /* Set file pointer to "offset". */
#define SEEK_CUR    1   /* Set file pointer to current plus "offset". */
#define SEEK_END    2   /* Set file pointer to EOF plus "offset". */
#endif

/*
 * Mode parameter for fopen() and spdFopen().  Added by S.Yamamoto 2000/02/10.
 */

#ifdef WIN32

#define SPD_FOPEN_RT_MODE   "rt"
#define SPD_FOPEN_RB_MODE   "rb"
#define SPD_FOPEN_WT_MODE   "wt"
#define SPD_FOPEN_WB_MODE   "wb"
#define SPD_FOPEN_AT_MODE   "at"
#define SPD_FOPEN_AB_MODE   "ab"

#else /* Almost all UNIXes. */

#define SPD_FOPEN_RT_MODE   "r"
#define SPD_FOPEN_RB_MODE   "rb"
#define SPD_FOPEN_WT_MODE   "w"
#define SPD_FOPEN_WB_MODE   "wb"
#define SPD_FOPEN_AT_MODE   "a"
#define SPD_FOPEN_AB_MODE   "ab"

#endif /* Almost all UNIXes. */

/******************************************************************************/

/*
 *                      SDB parameters.
 */

#define SAPID_TEXT_FORMAT_SDB   0
#define SAPID_BINARY_FORMAT_SDB 1

#define SPD_ENDIAN_UNKNOWN      (-1)
#define SPD_ENDIAN_BIG          0
#define SPD_ENDIAN_LITTLE       1

/* The third argument of function spdOpenSDBfile(). */
#define SAPID_O_READ            1
#define SAPID_O_WRITE           2
#define SAPID_O_APPEND          4

#include <Sapid/SapidDirectory.h>

/* 
 * SDB attribute separator.
 */
#define SDB_ATTR_SEPARATOR      '\1'

/* Max length of identifier in C language (not including terminater). */ 
#define SPD_ID_LEN              48

/*
 * Max length of constatnt string and comment (not including terminater).
 * Should check string literal and comment length in "lex.l" and "java-lex.l".
 * This value should be greater than FILENAME_MAX.
 */ 
/* #define SPD_STRING_LEN          (3 * 1024) */
#define SPD_STRING_LEN          (32 * 1024)

/*
 * Max length of a line or string attribute (not including terminater).
 */
#define SPD_LINE_LEN            SPD_STRING_LEN

/* 
 * Bit pattern of object id:
 * 
 * +------------------------- 32 bit -------------------------+
 * 
 * +--------------------+------------------+------------------+  
 * |sign part           |class id part     |object id part    |
 * +--------------------+------------------+------------------+  
 * 
 * + SPD_BIT_WIDTH_SIGN + SPD_BIT_WIDH_CID + SPD_BIT_WIDTH_ID +
 */
#define SPD_BIT_WIDTH_SIGN              1
#define SPD_BIT_WIDTH_CID               9
#define SPD_BIT_WIDTH_ID                22

#if (32 < SPD_BIT_WIDTH_SIGN + SPD_BIT_WIDTH_CID + SPD_BIT_WIDTH_ID)
    "Invalid configuration."
#endif

/* 
 * Object id in class which class id is classId:
 * SPD_OBJ_ID_OFFSET(classId) <= id < SPD_OBJ_ID_OFFSET(classId) + SPD_ID_SPACE
 */
#define SPD_CID_SPACE                   (1 << SPD_BIT_WIDTH_CID)
#define SPD_ID_SPACE                    (1 << SPD_BIT_WIDTH_ID)
#define SPD_OBJ_ID_OFFSET(classId)      ((classId) * SPD_ID_SPACE)
#define SPD_GET_CID_FROM_OBJ_ID(objId)  ((objId) / SPD_ID_SPACE)

/* For only spdGetIncludedOccInit(). */
#define SPD_ANY_FILE_ID     (SPD_NON_ID - 1)

/*
 * Macro SPD_NON_OFFSET is used as special value of invalid offset.
 * Offset starts from 0.
 */
#define SPD_NON_OFFSET      (-1)

/*
 * Macro SPD_NON_LENGTH is used as special value of invalid length of object.
 * Added by S.Yamamoto 1997/08/26.
 */
#define SPD_NON_LENGTH      (-1)

#define SPD_NON_LINENO      (-1)
#define SPD_NON_COLUMN      (-1)

#define SPD_NON_AID         (-1)

/* 
 * { fileId, objectId, offset, relationId }.
 */
#define SPD_NON_OCC \
    { SPD_NON_ID, SPD_NON_ID, SPD_NON_OFFSET, SPD_NON_ID }

/* 
 * { objectId, relationId }.
 */
#define SPD_NON_REL \
    { SPD_NON_ID, SPD_NON_ID }

/* 
 * { fileId, offset, length }.
 */
#define SPD_NON_REGION \
    { SPD_NON_ID, SPD_NON_OFFSET, SPD_NON_LENGTH }

/* 
 * { pblockId, offset, length }.
 */
#define SPD_NON_BLOCK \
    { SPD_NON_ID, SPD_NON_OFFSET, SPD_NON_LENGTH }

/* 
 * { size, *id }.
 */
#define SPD_EMPTY_ARRAY \
    { 0, NULL }

#define SAPID_CPP_NOT_SUPPORTED (-65)

#define SPD_DEFAULT_TARGET_PROGRAM_NAME     "a.out"

#define SPD_DEFAULT_TARGET_DOCUMENT_NAME    "Document"
#define SPD_DEFAULT_TARGET_SOFTWARE_NAME    "Software"

#define SPD_SKIPPED_BLOCK_IN_EXPRESSION     "__SPD_SKIPPED_BLOCK_IN_EXPRESSION"

#define SPD_FIRST_DB_ID         0       /* Temporarily added for AR4. */

/******************************************************************************/

/*
 *                      Miscellaneous macros.
 */

#define SPD_STREQ(x, y)         (strcmp((x), (y)) == 0)
#define SPD_STREQ3(x, y, z)     ((strcmp((x), (y)) == 0) || \
                                 (strcmp((x), (z)) == 0))
#define SPD_STREQ_N(x, y, n)    (strncmp((x), (y), (n)) == 0)

#ifdef WIN32
#define SPD_STREQ_ICASE(x, y)   (stricmp((x), (y)) == 0)
#else
#define SPD_STREQ_ICASE(x, y)   (strcasecmp((x), (y)) == 0)
#endif

#define SPD_MIN(x, y)           ((x) < (y) ? (x) : (y))
#define SPD_MAX(x, y)           ((x) < (y) ? (y) : (x))

#define SPD_QSORT(b, n, s, c) \
        spdQsortTmp((void *)(b), (size_t)(n), (size_t)(s), \
                    (int (*)(const void *, const void *))(c))

/* Replace Solaris's native function qsort(). */
#if defined(__sparc)
#define spdQsortTmp(b, n, s, c)     spdQsort((b), (n), (s), (c))
#else
#define spdQsortTmp(b, n, s, c)     qsort((b), (n), (s), (c))
#endif

/* 
 * #define SPD_BSEARCH(k, b, n, s, c) \
 *         bsearch((void *)(k), (void *)(b), (size_t)(n), (size_t)(s), \
 *                 (int (*)(const void *, const void *))(c))
 */

/*
 * If number of element in the table is zero, something wrong occurs in
 * some environments.
 *
 * IRIX 5.3: Segmentation fault
 * HP-UX 10.01: Infinite loop.
 */
#define SPD_BSEARCH(k, b, n, s, c) \
        ((n) != 0 ? \
         bsearch((void *)(k), (void *)(b), (size_t)(n), \
                 (size_t)(s), (int (*)(const void *, const void *))(c)) : \
         NULL)

#define SPD_ARRAY_SIZE(array)   (sizeof(array) / sizeof(array[0]))

/******************************************************************************/

/*
 *                      Basic functions.
 */

#include <Sapid/SpdUtil.h>

/******************************************************************************/

/*
 *                      HTML and XML definitions.
 */

/*
 * XML-declaration.
 */
#define SPD_XML_DECL_1_0 \
        "<?xml version=\"1.0\"?>\n"
/* #define SPD_XML_DECL_1_0_ISO_8859 \                         */
/*         "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n" */

/*
 * DOCTYPE.
 */
#define SPD_DOCTYPE_HTML_4_01 \
        "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n"
#define SPD_DOCTYPE_HTML_4_01_T \
        "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n"
#define SPD_DOCTYPE_HTML_4_01_F \
        "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\" \"http://www.w3.org/TR/html4/frameset.dtd\">\n"

#define SPD_DOCTYPE_XHTML_1_0_S \
        "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n"
#define SPD_DOCTYPE_XHTML_1_0_T \
        "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
#define SPD_DOCTYPE_XHTML_1_0_F \
        "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Frameset//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd\">\n"

/******************************************************************************/

/*
 *                      Access routines ver. [3-5] headers.
 */

#if defined(SAPID_AR3)
#include <Sapid/AR3.h>
#endif

#if defined(SAPID_AR4)
#include <Sapid/AR4.h>
#endif

#if defined(SAPID_AR5)
#include <Sapid/AR5.h>
#endif

#if defined(SAPID_AR4) || defined(SAPID_AR5)
#include <Sapid/Imodel.h>
#include <Sapid/Pmodel.h>
#include <Sapid/AreaModel.h>
#include <Sapid/Bmodel.h>
#include <Sapid/Dmodel.h>
#include <Sapid/SDAmodel.h>
#include <Sapid/StmtModel.h>
#include <Sapid/sdaarea.h>
#include <Sapid/TypeView.h>
#include <Sapid/CommentView.h>
#include <Sapid/FlowView.h>
#include <Sapid/FCD-view.h>
#include <Sapid/Jmodel.h>
#endif /* defined(SAPID_AR4) || defined(SAPID_AR5) */

/*
 *                      PIDB definitions.
 */

#if defined(SAPID_AR3_KERNEL)
#include <Sapid/PIDB.h> 
#endif

#include <Sapid/PIDBAR.h> 

/*
 *                      SMU definitions. 
 */

#if defined(SAPID_AR4) || defined(SAPID_AR5)
/* Note, "SMU.h" is included in "AR4.h". */ 
#include <Sapid/SMU.h>
#endif

/******************************************************************************/

/*
 *                      Error messages.
 */

#define SPD_MSG_NO_SDB          "\"%s\": No such SDB.\n"
#define SPD_MSG_INVALID_SDB     "\"%s\": Invalid SDB.\n"

/******************************************************************************/

/*
 *                      Debug facilities.
 */ 

/* Dirty trick for lint, "gcc -Wall" and VCC checking. */

#if defined(lint) || defined(WIN32)
#define SPD_DUMMY_RETURN(x)     return (x)
#else
#define SPD_DUMMY_RETURN(x)     /* Empty. */
#endif

/******************************************************************************/

#ifdef SAPID_SAFETY_MODE_0

int _spdGetAttrValInt(char *d_file, int d_line,
                      int objectId, SpdConstString attrName);
int _spdGetAttrValIntOfObj(char *d_file, int d_line,
                           int objectId, SpdString attrName);
int _spdGetAttrValIntOfRel(char *d_file, int d_line,
                           SpdRelId relationId, SpdConstString attrName);
SpdString _spdGetAttrValString(char *d_file, int d_line,
                               SpdId objectId, SpdString attrName);

#define spdGetAttrValInt(o, a) \
        _spdGetAttrValInt(__FILE__, __LINE__, (o), (a))
#define spdGetAttrValIntOfObj(o, a) \
        _spdGetAttrValInt(__FILE__, __LINE__, (o), (a))
#define spdGetAttrValIntOfRel(r, a) \
        _spdGetAttrValIntOfRel(__FILE__, __LINE__, (r), (a))
#define spdGetAttrValString(o, a) \
        _spdGetAttrValString(__FILE__, __LINE__, (o), (a))

#endif /* SAPID_SAFETY_MODE_0 */

/******************************************************************************/

/* 
 * Debug flag for read routines.
 */

/* #define SPD_DUMP_READ */

/* Debug for AR5's exaustive memory use 2002/10/06. */
/* #define SPD_DUMP_READ_1 */

/******************************************************************************/

#define SPD_MKSTR(x)            (# x)

/******************************************************************************/

/* 
 * !! Quick dirty hack for analyzing va_arg() with sdbin4 on Sparc architecture.
 */

#if defined(SPD_DUMMY_VA_ARG) && defined(__sparc)
#undef va_arg
#define va_arg(x, y)            (x)
#endif

#endif /* not _SAPID_H */
