/* 
 * Program:             $RCSfile: spdSet2.h,v $  $Revision: 30.8 $
 * 
 * Purpose:             
 * 
 * Author:              M.Hakamata  1998/07/05
 *                      A.Yoshida  2003/07/07
 *                      S.Yamamoto  1998/07/06
 * 
 * (C) Copyright:       M.Hakamata, A.Yoshida and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef _SPDSET2_H
#define _SPDSET2_H

/* Base type. */
#define SPDSET_HASH                     (0)
#define SPDSET_BTREE                    (1)
          
/* Key type. */
#define SPDSET_INT_KEY                  (0)
#define SPDSET_LONG_KEY                 (1)
#define SPDSET_STRING_KEY               (2)
#define SPDSET_OBJID_KEY                (3)
#define SPDSET_COPY_STRING_KEY          (4)
          
/* Value type. */
#define SPDSET_KEY_IS_VALUE             (0)
#define SPDSET_INT_VALUE                (1)
#define SPDSET_LONG_VALUE               (2)
#define SPDSET_POINTER_VALUE            (3)
#define SPDSET_STRING_VALUE             (4)
#define SPDSET_OBJID_VALUE              (5)
#define SPDSET_COPY_STRING_VALUE        (6)

#if defined(__STDC__) || defined(__cplusplus)
typedef void            *SpdSetAnyPointer;
#else
typedef int             *SpdSetAnyPointer;
#endif

typedef union {
    int                 word;
    long int            long_word;
    SpdObjId            objId;
    SpdSetAnyPointer    ptr;
} SpdElemValue;

typedef struct SpdElem_ {
    struct SpdElem_     *next;
    unsigned int        rawHash;
    SpdElemValue        value;
    SpdElemValue        key;
} SpdElem;

typedef struct SpdElemBlk_ {
    struct SpdElemBlk_  *prev;
    unsigned int        blkSize;
    unsigned int        usedNum;
    SpdElem             *array;
} SpdElemBlk;

struct SpdSet2__;

typedef struct {
  struct SpdSet2__      *set;
  unsigned int          curBkt;
  SpdElem               *curElem;
} SetCursor;

typedef unsigned int (* SpdSetHashF)(struct SpdSet2__ *set, SpdElemValue key);
typedef int (* SpdSetElemEqual)(SpdElemValue e1, SpdElemValue e2);
typedef void (* SpdSetElemFree)(SpdElemValue e);
typedef void (*SpdSetElemCopy)(SpdElemValue *dst, SpdElemValue *src);

typedef struct SpdSet2__ {
    int                 keyType;
    SpdSetHashF         keyHash;
    SpdSetElemEqual     keyEqual;   /* equality function of key elements */
    SpdSetElemFree      keyFree;    /* free function of key elements */
    SpdSetElemCopy      keyCopy;    /* deep copy of key elements */

    int                 valueType;
    SpdSetElemFree      valueFree;  /* free function of value elements */
    SpdSetElemCopy      valueCopy;  /* deep copy of value elements */

    unsigned long elemNum;       /* Number of elements. */

    unsigned long rebuildNum;    /* Rebuild hash table if elemNum gets to be
                                    larger more than this value. */
    SpdElem       **bktArray;    /* bucket array */
    size_t        bktNum;        /* Number of bucket. */
    size_t        initialBktNum; /* initial bucket number */

    SpdElemBlk    *pool;         /* pointer to list of allocated SpdElemBlk */
    unsigned long poolSize;
    unsigned long poolUsed;
    unsigned int  poolCapacityIncrement;
    unsigned int  poolInitialCapacity;
    
    SpdElem       *freeList;     /* pointer to list of freed SpdElem */
} SpdSet2_;

typedef SpdSet2_ *SpdSet2;

/* #define SPDSET2_MEMORY_DEBUG */

/* spdCreateSet() and spdFreeSet(). */

#ifdef SPDSET2_MEMORY_DEBUG
SpdSet2 spdCreateSet2Dbg(int keyType, int valueType, char *file, int line);
#define spdCreateSet2(key, value) \
        (spdCreateSet2Dbg(key, value, __FILE__, __LINE__))
#else
SpdSet2 spdCreateSet2(int keyType, int valueType);
#endif

void spdFreeSet2(SpdSet2 set);

/* spdContains(). */

SpdBoolean spdContains2(SpdSet2 set, SpdElemValue key);

/* Wrapper functions of spdContains2().
 *   S: String key.
 *   I: Integer key.
 *   L: Long interger key.
 *   O: Object identifier key, SpdObjId key.
 */

SpdBoolean spdContainsS(SpdSet2 set, SpdConstString key);
SpdBoolean spdContainsI(SpdSet2 set, int key);
SpdBoolean spdContainsL(SpdSet2 set, long key);
SpdBoolean spdContainsO(SpdSet2 set, SpdObjId key);

/* spdSearchElem(). */

SpdElem *spdSearchElem2(SpdSet2 set, SpdElemValue key);

/* Wrapper functions of spdSearchElem2(). */

SpdElem *spdSearchElemS(SpdSet2 set, SpdConstString key);
SpdElem *spdSearchElemI(SpdSet2 set, int key);
SpdElem *spdSearchElemL(SpdSet2 set, long key);
SpdElem *spdSearchElemO(SpdSet2 set, SpdObjId key);

/* spdDelElem(). */

SpdBoolean spdDelElem2(SpdSet2 set, SpdElemValue key);

/* Wrapper functions of spdDelElem2(). */

SpdBoolean spdDelElemS(SpdSet2 set, SpdConstString key);
SpdBoolean spdDelElemI(SpdSet2 set, int key);
SpdBoolean spdDelElemL(SpdSet2 set, long key);
SpdBoolean spdDelElemO(SpdSet2 set, SpdObjId key);

/* spdAddElem(). */

SpdElem *spdAddElem2(SpdSet2 set, SpdElemValue key, SpdElemValue val);

/*
 * Wrapper functions of spdAddElem2().
 *   K: Key as value.
 */
SpdElem *spdAddElemSK(SpdSet2 set, SpdConstString key);
SpdElem *spdAddElemSI(SpdSet2 set, SpdConstString key, int val);
SpdElem *spdAddElemSS(SpdSet2 set, SpdConstString key, SpdConstString val);
SpdElem *spdAddElemSP(SpdSet2 set, SpdConstString key, SpdSetAnyPointer val);
SpdElem *spdAddElemSL(SpdSet2 set, SpdConstString key, long val);
SpdElem *spdAddElemSO(SpdSet2 set, SpdConstString key, SpdObjId val);

SpdElem *spdAddElemIK(SpdSet2 set, int key);
SpdElem *spdAddElemII(SpdSet2 set, int key, int val);
SpdElem *spdAddElemIS(SpdSet2 set, int key, SpdConstString val);
SpdElem *spdAddElemIP(SpdSet2 set, int key, SpdSetAnyPointer val);
SpdElem *spdAddElemIL(SpdSet2 set, int key, long val);
SpdElem *spdAddElemIO(SpdSet2 set, int key, SpdObjId val);

SpdElem *spdAddElemLK(SpdSet2 set, long key);
SpdElem *spdAddElemLI(SpdSet2 set, long key, int val);
SpdElem *spdAddElemLS(SpdSet2 set, long key, SpdConstString val);
SpdElem *spdAddElemLP(SpdSet2 set, long key, SpdSetAnyPointer val);
SpdElem *spdAddElemLL(SpdSet2 set, long key, long val);
SpdElem *spdAddElemLO(SpdSet2 set, long key, SpdObjId val);

SpdElem *spdAddElemOK(SpdSet2 set, SpdObjId key);
SpdElem *spdAddElemOI(SpdSet2 set, SpdObjId key, int val);
SpdElem *spdAddElemOS(SpdSet2 set, SpdObjId key, SpdConstString val);
SpdElem *spdAddElemOP(SpdSet2 set, SpdObjId key, SpdSetAnyPointer val);
SpdElem *spdAddElemOL(SpdSet2 set, SpdObjId key, long val);
SpdElem *spdAddElemOO(SpdSet2 set, SpdObjId key, SpdObjId val);

/* Cursor. */

SetCursor spdSetCursorCreate(SpdSet2 set);
SpdBoolean spdSetCursorNext(SetCursor *csr);

/* Access functions to an element. */

int spdSetGetValueI(SpdElemValue elem);
long spdSetGetValueL(SpdElemValue elem);
SpdConstString spdSetGetValueS(SpdElemValue elem);
SpdObjId spdSetGetValueO(SpdElemValue elem);
SpdSetAnyPointer spdSetGetValueP(SpdElemValue elem);

void spdSetSetValueI(SpdElemValue *elem, int i);
void spdSetSetValueL(SpdElemValue *elem, long i);
void spdSetSetValueS(SpdElemValue *elem, SpdConstString s);
void spdSetSetValueO(SpdElemValue *elem, SpdObjId id);
void spdSetSetValueP(SpdElemValue *elem, SpdSetAnyPointer ptr);

/* Free functions. */

void spdElemFreeS(SpdElemValue elem);
void spdElemFreeP(SpdElemValue elem);
void spdElemFreeSpdRel(SpdElemValue elem);
void spdElemFreeSpdOcc(SpdElemValue elem);

/* Copy methods for elements. */

void spdElemCopyS(SpdElemValue *dst, SpdElemValue *src);
void spdElemMoveS(SpdElemValue *dst, SpdElemValue *src);
void spdElemCopySpdRel(SpdElemValue *dst, SpdElemValue *src);
void spdElemMoveSpdRel(SpdElemValue *dst, SpdElemValue *src);
void spdElemCopySpdOcc(SpdElemValue *dst, SpdElemValue *src);
void spdElemMoveSpdOcc(SpdElemValue *dst, SpdElemValue *src);

/* Misc. */

SpdSet2 spdDupSet2(SpdSet2 set);
SpdSet2 spdUnionSet2(SpdSet2 s1, SpdSet2 s2);
SpdSet2 spdAddSet2(SpdSet2 s1, SpdSet2 s2);
SpdSet2 spdIntersectionSet2(SpdSet2 s1, SpdSet2 s2);
SpdSet2 spdDifferenceSet2(SpdSet2 s1, SpdSet2 s2);
SpdSet2 spdSubtractSet2(SpdSet2 s1, SpdSet2 s2);

#endif /* _SPDSET2_H */
