/* 
 * Program:             $RCSfile: spdTraverse.h,v $  $Revision: 30.6 $
 * 
 * Purpose:             spdTraverse.
 * 
 * Author:              H.Suzuki  1998/02/04
 *                      S.Yamamoto  1998/02/08
 * Modified:            N.Iwamoto  1998/06/25
 * 
 * (C) Copyright:       H.Suzuki, N.Iwamoto and S.Yamamoto  1998
 *                      This file is a product of the project Sapid.
 */

#define SAPID_TRV_GO        0 
#define SAPID_TRV_STOP      1
#define SAPID_TRV_SKIP      2
#define SAPID_TRV_PAUSE     3

typedef enum {
    SAPID_TRV_ACT_BEFORE = 0,
    SAPID_TRV_ACT_AFTER
} SpdTrvActType;

typedef enum {
    SAPID_TRV_CALLED_NOTHING = 0,
    SAPID_TRV_CALLED_BEFORE_REL,
    SAPID_TRV_CALLED_BEFORE_OBJ,
    SAPID_TRV_CALLED_AFTER_OBJ,
    SAPID_TRV_CALLED_AFTER_REL,
    SAPID_TRV_CALLED_MID_REL
} SpdTrvActLevel;

typedef struct {
    size_t size;
    size_t elemSize;
    void **elem;
} SpdStack;

typedef struct {
    SpdObjId objectId, pObjId;
    SpdRel rel;
    SpdTrvActType type;
    SpdTrvActLevel actLv;
    int depth, num;
} SpdTrvStackElem;

typedef struct {
    SpdSet classBeforeRelFuncSet;
    SpdSet classAfterRelFuncSet;
    SpdSet classBeforeFuncSet;
    SpdSet classMiddleFuncSet;
    SpdSet classAfterFuncSet;
    SpdSet objBeforeFuncSet;
    SpdSet objAfterFuncSet;
    SpdSet sortFuncSet;
    SpdSet trvRelSet;
    SpdStack *trvStack;
} SpdTrvActInfo;

typedef int (*SpdTrvActRelFunc)(SpdObjId, SpdRel, void *);
typedef int (*SpdTrvActMidFunc)(SpdObjId, int, void *);
typedef int (*SpdTrvActFunc)(SpdObjId, void *);

void spdTraverse(SpdObjId id, SpdTrvActInfo *info, void *vp);
void spdTraverseContinue(SpdTrvActInfo *info, void *vp);
void spdTrvAddRel(SpdTrvActInfo *, SpdClassId);
void spdTrvAddSortFunc(SpdTrvActInfo *, SpdClassId, SpdCompareRelFunc);
SpdTrvActRelFunc spdTrvAddActFuncOfRelClass(SpdTrvActInfo *, SpdClassId,
                                            SpdTrvActType, SpdTrvActRelFunc);
SpdTrvActFunc spdTrvAddActFuncOfClass(SpdTrvActInfo *, SpdClassId,
                                      SpdTrvActType, SpdTrvActFunc);
SpdTrvActMidFunc spdTrvAddActFuncMidOfClass(SpdTrvActInfo *, SpdClassId,
                                            SpdTrvActMidFunc);
SpdTrvActFunc spdTrvAddActFuncOfObj(SpdTrvActInfo *,
                                    SpdObjId, SpdTrvActType, SpdTrvActFunc);
SpdTrvActInfo *spdTrvInitActInfo(void);
void spdTrvFreeActInfo(SpdTrvActInfo *);
SpdStack *spdTrvCreateStack(void);
